/* ------------------------------------------------------------------------
 * $Id: StackInvocation.java 25 2010-10-02 08:40:02Z tpv $
 * Copyright 2007, Tim Vernum
 *
 * Permission is hereby granted, free of charge, to any person obtaining 
 * a copy of this software and associated documentation files (the 
 * "Software"), to deal in the Software without restriction, including 
 * without limitation the rights to use, copy, modify, merge, publish, 
 * distribute, sublicense, and/or sell copies of the Software, and to 
 * permit persons to whom the Software is furnished to do so, subject 
 * to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ------------------------------------------------------------------------
 */

package net.sf.tie;

import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Method;
import java.util.Iterator;

import net.sf.tie.internal.MethodFinder;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;

/**
 * The concrete implementation of the <a href="http://aopalliance.sf.net">AOP Alliance</a>
 * interface {@link MethodInvocation}, and the <i>Tie</i> interface {@link WrappedMethodInvocation}.
 * This invocation walks through a collection of interceptors, and then calls a method on a supplied object.
 * <b>This class is not part of the public API for Tie</b>
 */
class StackInvocation implements MethodInvocation, WrappedMethodInvocation
{
    private final Iterator<MethodInterceptor> _iterator;
    private final Object _endPoint;
    private final MethodCall _call;
    private final Method _implementationMethod;

    public StackInvocation(Iterable<MethodInterceptor> interceptors, Object endPoint, MethodCall call)
    {
        _iterator = interceptors.iterator();
        _endPoint = endPoint;
        _call = call;
        _implementationMethod = lookupImplementation();
    }

    private Method lookupImplementation()
    {
        Class< ? extends Object> implementationClass = _endPoint.getClass();
        Method interfaceMethod = _call.getMethod();

        Method method = MethodFinder.matchMethod(implementationClass, interfaceMethod);
        if (method == null)
        {
            throw new IllegalStateException("Cannot find method " + interfaceMethod + " on " + implementationClass);
        }
        return method;
    }

    public Method getMethod()
    {
        return _call.getMethod();
    }

    public Method getMethodDeclaration()
    {
        return _call.getMethod();
    }

    public Method getMethodImplementation()
    {
        return _implementationMethod;
    }

    public Object[] getArguments()
    {
        return _call.getArguments();
    }

    public Object proceed() throws Throwable
    {
        if (_iterator.hasNext())
        {
            return _iterator.next().invoke(this);
        }
        else
        {
            return _call.invoke(_endPoint);
        }
    }

    public Object getThis()
    {
        return _endPoint;
    }

    public AccessibleObject getStaticPart()
    {
        return getMethod();
    }

    public String toString()
    {
        return getClass().getName() + "{" + _call + " on " + _endPoint + "}";
    }
}
